#include "pest.h"
#include "addr.h"
#include "peer.h"
#include "packet.h"
#include "crypto.h"
#include "serial.h"
#include "log.h"
#include "time.h"
#include "util.h"
#include "state.h"
#include <assert.h>

static void ac_black_ciph(uint8_t        *ab,
                          uint8_t const  *ar,
                          uint8_t const  *key_serp){

  serpent_cbc_ciph(key_serp,ab+AC_RED_OFF,ar,AC_RED_SIZE);
}

static void ac_black_deciph(uint8_t        *ar,
                            uint8_t const  *ab,
                            uint8_t const  *key_serp){

  serpent_cbc_deciph(key_serp,ar,ab+AC_RED_OFF,AC_RED_SIZE);
}

static void ac_black_sign(uint8_t        *ab,
                          uint8_t const  *key){

  hmac_sha384(ab+AC_SIGN_OFF,key,KEY_SIGN_SIZE,ab+AC_RED_OFF,AC_RED_SIZE);
}

static bool ac_black_verify(uint8_t const  *ab,
                            uint8_t const  *key){

  uint8_t sign[AC_SIGN_SIZE];
  hmac_sha384(sign,key,KEY_SIGN_SIZE,ab+AC_RED_OFF,AC_RED_SIZE);
  return memcmp(sign,ab+AC_SIGN_OFF,AC_SIGN_SIZE) == 0;
}

static bool ac_send(uint8_t const *ar,peer_t *p){
  uint8_t   ab[AC_BLACK_SIZE];
  pkey_t   *pk;
  uint64_t  stamp;

  pk = peer_key_recent(p);
  assert(pk);

  ac_black_ciph(ab,ar,pk->ciph_serp);
  ac_black_sign(ab,pk->key.sub.sign);

  if(!rng_read(rng_fast_fd,ab+AC_BPAD_OFF,AC_BPAD_SIZE)){
    return 0;
  }

  stamp = mono_sec();

  assert(sizeof(ab) == TEXT_SIZE);
  return send_text(CMD_ADDRESSCAST,stamp,ab,sizeof(ab),0,0);
}

static bool ac_red_write(uint8_t *ar,addr_t const *addr){
  uint32_t cmd  = 0;
  uint16_t port = addr->port;
  uint32_t ip   = byteswap32(addr->ip);

  if(!rng_read(rng_fast_fd,ar+AC_NONCE_OFF,AC_NONCE_SIZE)){
    return 0;
  }

  write_u32(ar+AC_CMD_OFF,  &cmd);
  write_u16(ar+AC_PORT_OFF, &port);
  write_u32(ar+AC_IP_OFF,   &ip);

  if(!rng_read(rng_fast_fd,ar+AC_RPAD_OFF,AC_RPAD_SIZE)){
    return 0;
  }

  return 1;
}

static bool ac_red_read(uint8_t const *ar,addr_t *addr){
  uint32_t cmd;
  uint16_t port;
  uint32_t ip;

  read_u32(ar+AC_CMD_OFF,  &cmd);
  read_u16(ar+AC_PORT_OFF, &port);
  read_u32(ar+AC_IP_OFF,   &ip);

  if(cmd != 0) return 0;

  addr->port  = port;
  addr->ip    = byteswap32(ip);
  return 1;
}

void addrcast_broadcast(peer_t *p){
  uint8_t ar[AC_RED_SIZE];

  if(addr_null(&acaddr)) return;

  if(!ac_red_write(ar,&acaddr)){
    irclog_err("failed to create address cast");
    return;
  }

  if(p){
    if(!p->active &&
       !p->paused &&
       p->key_cnt > 0){
      ac_send(ar,p);
    }
  }else{
    for(size_t x = 0;x < peer_cnt;x++){
      peer_t *p = peer+x;
      if(p->active)       continue;
      if(p->paused)       continue;
      if(p->key_cnt == 0) continue;
      ac_send(ar,p);
    }
  }

  BUF_ZERO(ar);
}

void addrcast_periodic(uint64_t time){
  if(addrcast_last == 0) goto ret;
  if(time < addrcast_last+acint) return;

  addrcast_broadcast(0);

ret:
  addrcast_last = time;
}

static bool find_key(uint8_t const *ab,peer_t **p_out,pkey_t **pk_out){
  uint16_t peer_idx[peer_cnt];

  if(!index_shuffle(peer_idx,peer_cnt)){
    irclog_err("could not shuffle index, "
               "address cast cannot be processed");
    return 0;
  }

  for(size_t x = 0;x < peer_cnt;x++){
    peer_t *p = peer+peer_idx[x];

    if(p->paused) continue;
    if(p->active) continue;

    for(size_t y = 0;y < p->key_cnt;y++){
      pkey_t *pk = p->key+y;
      if(!ac_black_verify(ab,pk->key.sub.sign)) continue;
      *p_out  = p;
      *pk_out = pk;
      return 1;
    }
  }
  return 0;
}

//ret 1 if im recipent
static bool ac_handle(uint8_t const *ab,char const *speaker){
  uint8_t   ar[AC_RED_SIZE];
  peer_t   *p;
  pkey_t   *pk;
  addr_t    addr;

  if(!find_key(ab,&p,&pk)){
    return 0;
  }

  if(peer_lookup_noerr(speaker) != p){
    log_err("speaker mismatch in address cast received from peer '%s'",
            p->handle[0]);
    return 1;
  }

  ac_black_deciph(ar,ab,pk->ciph_serp);
  if(!ac_red_read(ar,&addr)){
    log_err("nonzero address cast command received from peer '%s'",
            p->handle[0]);
    return 1;
  }
  BUF_ZERO(ar);

  if(!addr_equal(&p->addr,&addr)){
    peer_addr_set(p,&addr,0);
    prod_send(p,PROD_FLAG_REQANS);
  }
  return 1;
}

void addrcast_handle(uint8_t       *r,
                     hash_t const   hash,
                     uint8_t        bounces,
                     char const    *speaker){

  uint8_t const *m  = r+MESSAGE_OFF;
  uint8_t const *ab = m+TEXT_OFF;

  if(!ac_handle(ab,speaker)){
    red_rebroadcast(r,hash,bounces);
  }
}

